!
! Module: bkgoptrate_aux
!
! Authors:  Masaaki Sakano (SSC/LUX)
!          
! Description: Core subroutines/functions for bkgoptrate.
! 
! Comments: 
! 
! $Id: bkgoptrate_aux.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module bkgoptrate_aux

  use dal	! For type definition of single etc.
  ! use param
  use errorHandling
  ! use datafilter

  ! use array_utils	! for smallest/largestTrueIndex

  implicit none 

  private 

  public getSN2ofTs
  public getSN2ofTsOfRate
  public findOptimumY
  public calcThresholdFromAny
  public calcThresholdFromRate
  public calcThresholdFromCount

  type, public :: InfoFracexpT
    character(20) :: style = ''	! (calc|threshold|none)
      ! calc: Calculate the count or rate according to the FRACEXP
      ! threshold: Ignore the bins with FRACEXP lower than the threshold
    logical :: doUse = .false.
    character(64) :: colName = ''	! 'FRACEXP'
    real(single)  :: lowerThreshold = 0.0	! valid if style=='threshold'
    real(double), allocatable  :: ary(:)
  end type InfoFracexpT

  type, public :: ArgTimesStartEndT
    logical :: withStart=.false., withEnd=.false.
    real(single)  :: start=0.0, end=0.0
  end type ArgTimesStartEndT

  ! interface getSN2ofTs
  !   module procedure getSN2ofTsWithoutExp, getSN2ofTsWithExp
  ! end interface

contains 

!   !---------------------------------------------------------------------
!   function getSN2ofTsWithoutExp(tSeries, mask) result(sn2)
!   !---------------------------------------------------------------------
!   ! Returns the square of S/N for the given time-series.
! 
!     implicit none
! 
!     real(single) :: sn2
!     real(single), intent(in) :: tSeries(:)
!     logical,      intent(in), optional :: mask(size(tSeries))
! 
!     !-------------------------------------------------------------------
!     ! Start
! 
!     if (present(mask)) then
!       ! sn2 = count(mask)   / sqrt(sum(tSeries, MASK=mask))
!       sn2 = real(count(mask))**2   / sum(tSeries, MASK=mask)	! real() is needed to avoid a potential overflow in Integer --- anyhow conversion to Real would be internally needed to give the Real result to match with the 2nd clause (sum of Real).
!     else
!       sn2 = real(size(tSeries))**2 / sum(tSeries)
!     end if
! 
!   end function getSN2ofTsWithoutExp
! 
!   !---------------------------------------------------------------------
!   function getSN2ofTsWithExp(tSeries, aryExp, isRate, mask) result(sn2)
!   !---------------------------------------------------------------------
!   ! Returns the square of S/N for the given time-series.
! 
!     implicit none
! 
!     real(single) :: sn2
!     real(single), intent(in) :: tSeries(:)
!     real(single), intent(in) :: aryExp(size(tSeries))
!     logical,      intent(in) :: isRate
!     logical,      intent(in), optional :: mask(size(tSeries))
! 
!     real(single) :: localAryExp(size(tSeries)), denominator
!     integer(int32) :: i
! 
!     !-------------------------------------------------------------------
!     ! Start
! 
!     if (isRate) then
!       localAryExp = aryExp
!       denominator = 0.0
!       if (present(mask)) then
!         do i=1, size(tSeries)
!           denominator = denominator + aryExp(i)      * tSeries(i)
!         end do
!       else
!         where(.not. mask) localAryExp = 0.0	! Propagates the input mask.
!         do i=1, size(tSeries)
!           denominator = denominator + localAryExp(i) * tSeries(i)
!         end do
!       end if
! 
!       sn2 = sum(localAryExp)**2        / denominator
!     else
!       sn2 = sum(aryExp, MASK=mask)**2  / sum(tSeries, MASK=mask)
!     end if
! 
!   end function getSN2ofTsWithExp

  !---------------------------------------------------------------------
  function getSN2ofTs(tSeries, mask, aryExp) result(sn2)
  !---------------------------------------------------------------------
  ! Returns the square of S/N for the given time-series (of Count),
  ! for a source with no time variability.
  ! Strictly speaking, the true S/N is (SourceCount * getSN2ofTs()).

    implicit none

    real(single) :: sn2
    real(single), intent(in) :: tSeries(:)
    logical,      intent(in), optional :: mask(size(tSeries))
    real(single), intent(in), optional :: aryExp(size(tSeries))

    !-------------------------------------------------------------------
    ! Start

    if (present(aryExp)) then
      sn2 = sum(aryExp, MASK=mask)**2  / sum(tSeries, MASK=mask)
    else
      if (present(mask)) then
        sn2 = real(count(mask))**2   / sum(tSeries, MASK=mask)	! real() is needed to avoid a potential overflow in Integer --- anyhow conversion to Real would be internally needed to give the Real result to match with the 2nd clause (sum of Real).
      else
        sn2 = real(size(tSeries))**2 / sum(tSeries)
      end if
    end if

  end function getSN2ofTs

  !---------------------------------------------------------------------
  function getSN2ofTsOfRate(tSeries, mask, aryExp) result(sn2)
  !---------------------------------------------------------------------
  ! Returns the square of S/N for the given time-series of Rate.
  ! 
  ! NOTE: It is probably better to convert the time-series of rate
  !  to that of count beforehand and call getSN2ofTs() instead,
  !  considering the overhead to call this routine repeatedly.
  !   For that reason, this subroutine has not been tested!  Treat with care...

    implicit none

    real(single) :: sn2
    real(single), intent(in) :: tSeries(:)
    logical,      intent(in), optional :: mask(size(tSeries))
    real(single), intent(in), optional :: aryExp(size(tSeries))

    real(single) :: localAryExp(size(tSeries)), denominator
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    localAryExp = aryExp
    denominator = 0.0
    if (present(mask)) then
      where(.not. mask) localAryExp = 0.0	! Propagates the input mask.
    end if

    do i=1, size(tSeries)
      denominator = denominator + localAryExp(i) * tSeries(i)
    end do

    sn2 = sum(localAryExp)**2 / denominator
    ! NOTE: I am not sure if NaN is treated all right without causing an error.

  end function getSN2ofTsOfRate

  !---------------------------------------------------------------------
  subroutine findOptimumY(tsCnt, optimumY &
     , optimumSize, tsRate, aryFracexp, mask, lowerNbinRatio, status &
     , doDump)
  !---------------------------------------------------------------------
  ! *** Description ***
  !  Find the optimum cut of Y-value (Count or Rate) that gives the maximum
  ! S/N ratio for the input time-series, which should be the background
  ! time-series, where
  !    S/N =~ Tot_EXP / SQRT(Tot_Cnt)  (x Rate (=Constant) to give the counts).
  !  Then you (the calling routine) can exclude those high-background periods
  ! from another source time-series, based on the optimum background cut.
  ! 
  ! *** Arguments ***
  !  tsCnt(:) : Array of Counts for time bin.
  !    This is used to estimate the S/N.
  !    Note the unit has to be count (or alike).  Threrefore, if the original
  !    time-series is in unit of Rate, tsCnt(:) has to be calculated and
  !    given to this subroutine by the caller.
  !  optimumY   : The calculated optimum Y-value.
  !  optimumSize : The size of tsCnt(:) when it gives the best S/N (=optimumY).
  !    This is equivalent to count(mask) (after mask is modified in this routine).
  !  tsRate(:) : Array of Rate for each time bin (OPTIONAL).  If not given,
  !    (1) if aryFracexp is not given either, tsRate is set to be tcCnt,
  !    (2) if aryFracexp given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !    This is used to estimate the optimum threshold.
  !  aryFracexp(:) : Array of FRACEXP (OPTIONAL).  If not given,
  !    (1) if tsRate is not given either, aryFracexp(:) is set to be all 1.0,
  !    (2) if tsRate given, it is calculated accordingly.
  !    tsRate and aryFracexp should not be independent, but this subroutine
  !    does not check consistency when both the arguments are given.
  !  mask(:)    : Mask Array, indicating which indices in tsCnt(:) are valid.
  !    This mask is modified in the subroutine, and the returned mask shows
  !    true for all the valid values when the optimum value is found.
  !  lowerNbinRatio : Threshold in ratio for the number of bin used,
  !    where the calculation stops.  For example, if it is 0.05,
  !    this routine does not search for the optimum Y-value beyond the point,
  !    where the number of bins used is 5% of the entire bin number
  !    after the given mask is applied; that is, for example,
  !    when the size(tsCnt) is 1000 and the given mask has 80% of .true.,
  !    this routine does not search for the optimum value for the number
  !    of bins used being less than 40 (=1000*0.8*0.05), and if it happens,
  !    either error('noThresholdFound') is raised or status=-2 is returned.
  !    The default is 0.0.
  !  status     : Result (fitting) status indicator (OPTIONAL).
  !    0 for normal end,
  !    -2 if the optimum values is not found even when the search reaches the end
  !    (in that case, if status is not given, Error is raised.).
  !  doDump     : For debugging - it dumps all the S/Ns if True.
  !    Note the returned values are most likely to be incorrect if True.
  ! 
  ! *** NOTE ***
  !  At the time of writing, the minimization (maximization) algorithm
  ! in this subroutine is extremely crude and inefficient --- should
  ! be rewritten with the use of a proper minimization routine, ideally.
  ! 
  ! *** ALGORITHM ***
  !   0. Prepare TSeries(:), TRate(:) and Mask(:).
  !   1. Calculate the S/N in TSeries, based on Mask.
  !   2. Find the maximum value and its locations, namely the indices, in TRate(:).
  !   3. Set Mask(:) false for the index(ices), which have the maximum value,
  !  and calculate the S/N in TSeries, based on (revised) Mask.
  !   4. If the new S/N is larger than the previous S/N,
  !  then repeat the processing from (2).  If smaller,
  !  then the previous maximum value and Mask are the optimum ones (best S/N),
  !  so return them.
  ! 

    implicit none

    real(single),   intent(in) :: tsCnt(:)
    real(single),   intent(out) :: optimumY
    integer(int32), intent(out), optional :: optimumSize
    real(single),   intent(in),  optional :: tsRate(size(tsCnt)), aryFracexp(size(tsCnt))
    ! logical,        intent(in),  optional :: isRate	! This has to be provided if aryExp(:) is given.
    logical,        intent(inout), optional :: mask(size(tsCnt))
    ! logical,        intent(out), optional :: maskOut(size(tsCnt))
    real(single),   intent(in),  optional :: lowerNbinRatio
    integer,        intent(out), optional :: status
    logical,        intent(in),  optional :: doDump

    logical :: prevMask(size(tsCnt)), curMask(size(tsCnt)), isFracexpUsed, localDoDump
    real(single) :: localTsRate(size(tsCnt)), aryExp(size(tsCnt)) &
       , prevSN, curSN, prevprevY, prevY, curY
    integer(int32) :: nbin, sizeMaskIn, lowerNbin, nDup
    integer :: arMaxloc(1), prevLoc, curLoc

logical, parameter :: localIsTest = .false.

    !-------------------------------------------------------------------
    ! Start

    ! *** Algorithm (see also the above comment for the overview) ***
    !  Assumption: The function to give S/N is monotonic, namely having
    ! only one maximum value.
    !
    !  (1) Set False at the X (prevLoc) for the maximum Y (Rate) in the current mask (curMask).
    !  (2) Find the index X for the maximum Y (Rate) with the new mask (curLoc),
    !    as well as the maximum Y-value (curYRate).
    !  (3a) If curY==prevY then repeat from (1) after resetting prevLoc as curLoc.
    !  (3b) If not, proceed below.
    !  (4) Calculate S/N (curSN) for TS with the new mask (curMask).
    !  (5a) If (curSN > prevSN) then repeat from (1) after resetting prev*.
    !  (5b) If not, then returns prev* (nb. the returned Y-value is slightly larger than prevYRate)
    !  (6) If fails to find the maximum for SN, then it returns status=-2 or raises an error.
    !    However, even in that case, if over one-third of the lowest bins
    !    have the same Y(Rate)-value, AND if the lowerNbinRatio given is the minimum,
    !    then the Y(Rate)-value (or strictly speaking a value slightly higher
    !    than that) is returned and this subroutines ends normally.
    !    This case may happen in, such as, uniform time-series or step-function-like ones.
    !
    ! *** Notation ***
    !  (prev|cur)Mask : The mask file used to calculate the values.
    !  (prev|cur)SN   : Previous/Current SN when the above mask is applied.
    !  (prev|cur)Y    : Maximum Y(Rate)-value (when the above mask is applied).
    !  (prev|cur)Loc  : Its index             (when the above mask is applied).
    !  prevprevY      : Previous previous maximum Y(Rate)-value.

    optimumY = -1
    localTsRate(:) = tsCnt
    aryExp(:) = 1.0
    isFracexpUsed = .false.
    if (present(status)) status = 0
    if (present(optimumSize)) optimumSize = -1
    if (present(tsRate)) then
      isFracexpUsed = .true.
      localTsRate = tsRate
      if (present(aryFracexp)) then
        aryExp(:) = aryFracexp
      else
        where(localTsRate .ne. 0) aryExp(:) = tsCnt/localTsRate
        ! It is impossible to guess aryExp() when tsRate (tsCnt) is zero.
      end if
    else if (present(aryFracexp)) then
      isFracexpUsed = .true.
        aryExp(:) = aryFracexp
        where(aryExp .ne. 0) localTsRate(:) = tsCnt/aryExp
        ! It makes no sense to determine tsRate when aryExp is zero.
    end if
    ! if (present(maskOut)) maskOut = .false.
    localDoDump = .false.
    if (present(doDump)) localDoDump = doDump

    curMask = .true.
    if (present(mask)) curMask = mask
    sizeMaskIn = count(curMask)
    prevMask = curMask
    if (isFracexpUsed) then
      ! prevSN = getSN2ofTs(tsCnt, aryExp, .false., prevMask)
      prevSN = getSN2ofTs(tsCnt, prevMask, aryExp)
    else
      prevSN = getSN2ofTs(tsCnt, prevMask)
      ! prevSN = sizeMaskIn / sqrt(sum(tsCnt, MASK=prevMask))	! S/N calculation.
    end if

    if (present(lowerNbinRatio)) then
      lowerNbin = nint(sizeMaskIn * lowerNbinRatio)
      if ((lowerNbin == 0) .or. (lowerNbin == 1)) then
        lowerNbin = 2	! when lowerNbinRatio=0.0
      else if (lowerNbin < 0) then
        ! lowerNbinRatio has to be zero or positive.
        call error('Bug_lowerNbinRatio', 'Contact the code developer.')
      end if
    else
      lowerNbin = 2
    end if

    if (verbosity(AppMsg) >= Noisy) then
      if (present(lowerNbinRatio)) then
        write(errStr, '(a,i8,a,i8,a,i6,a,f7.4,a)') "Start findOptimumY(): Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=", lowerNbinRatio &
           , "  which will be applied further)."
      else
        write(errStr, '(a,i8,a,i8,a,i6,a)') "Start findOptimumY():  Size TS(" &
           , size(tsCnt), "), mask(", sizeMaskIn &
           , "), lowerNbin=", lowerNbin, " (lowerNbinRatio=UNSPECIFIED)."
      end if
      call message(AppMsg, Noisy)
    end if

    if (localIsTest) then
      print *,"DEBUG: nbin=",count(prevMask),"curSN2=",prevSN,"in Start."
      ! print *,"DEBUG: nbin=",count(curMask),"curSN2=",prevSN,"in Start."
    end if

    arMaxloc = maxloc(localTsRate, prevMask)
    prevLoc = arMaxloc(1)
    prevY = localTsRate(prevLoc)	! == maxval(localTsRate, prevMask)
    ! prevY = prevY + abs(prevY)/100.0	! slightly larger than the first maximum value.
    ! prevY = prevY + abs(epsilon(prevY)) * 4	! slightly larger than the first maximum value.
    prevprevY = prevY 
    nDup = 1

    do nbin=sizeMaskIn, lowerNbin, -1	! (lowerNbin >= 2) is guaranteed.
      ! nbin is the number of bins to be used in the estimate in each cycle.

      curMask(prevLoc) = .false.
      arMaxloc = maxloc(localTsRate, curMask)	! To calculate curLoc.
      curLoc = arMaxloc(1)
      curY = localTsRate(curLoc)	! Maximum Y for curMask

      if (prevY == curY) then	! The current maximum Y-value is identical to the previous Y-value, so should be taken away before the next evaluation.
        prevLoc = curLoc 
        nDup = nDup + 1
        if (localIsTest) then
          print *,"DEBUG: nbin=",count(curMask),"                    Removed-(i,Y)=",prevLoc,prevY
        end if
        cycle
      end if

      if (isFracexpUsed) then
        !curSN = getSN2ofTs(tsCnt, aryExp, .false., curMask)
        curSN = getSN2ofTs(tsCnt, curMask, aryExp)
        ! NOTE: This formula can replace the one in else-clause, in other words,
        !  this IF clause itself is not essential, however it would be slower.
      else
        curSN = getSN2ofTs(tsCnt, curMask)
      end if

      !THIS NOTE IS INCORRECT!!!!!
      ! NOTE: Here, count(curMask) should be equal to (nbin-1).
      !  Therefore the following equation should do the same job,
      !  and would be faster due to no overhead.
      ! curSN = (nbin-1) / sqrt(sum(tsCnt, MASK=curMask))
      !
      !  But currently I use the above function just to be conservative and safe.

      if (localIsTest) then
        print *,"DEBUG: nbin=",count(curMask),"curSN2=",curSN,"Removed-(i,Y)=",prevLoc,prevY
      end if
      if (curSN > prevSN) then
        nDup = 1
        prevSN = curSN
        prevprevY = prevY 
        prevY  = curY
        prevMask = curMask
        prevLoc = curLoc
        ! next (in loop)	! Make sure no practical statement is added after this!!!
      else
        ! The previous attempt gives the better S/N,
        ! so it is regarded as the best S/N value and is returned.
        write(errStr, '(a,es16.6,es16.6)') "S/N of previous/current are " &
           , prevSN, curSN
        call message(AppMsg, Verbose)
        write(errStr, '(a,i8,a,i8)') "The best S/N is achieved at i=" &
           , prevLoc, " nbin=",count(prevMask)
           !, curLoc, " nbin=",count(curMask)+nDup
        call message(AppMsg, Noisy)

        optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
        ! NOTE: The former (with a smaller coefficient, maybe 1) may be enough,
        !  or at least it is definitely enough, numerically speaking.
        !  However a simple x+epsilon(x) can be too small
        !  to distinguish in STDOUT by human eyes.
        ! On the other hand, equally, the latter would be probably enough.
        !  However if mask is specified for this routine, then the latter
        !  can be very large, whereas the user may not want such a large
        !  value for the later use, for example, when the user uses the value
        !  as an input for SAS-task tabgtigen, where the mask that the user
        !  specified for this task may not be taken into account.

!print *,"nDup=",nDup,", count(curMask)=",count(curMask),", count(prevMask)=",count(prevMask) 
        if (present(optimumSize)) optimumSize = count(prevMask)
        !if (present(optimumSize)) optimumSize = count(curMask) + nDup	! Plus nDup because the current removal of the index must be invalid and restored.
        if (present(mask)) then
           mask = prevMask
           ! maskOut = prevMask
        end if

        if (localDoDump) then
          ! Do nothing.
          nDup = 1
          prevSN = curSN
          prevprevY = prevY 
          prevY  = curY
          prevMask = curMask
          prevLoc = curLoc
        else
          return	! Found the solution!
        end if
      end if
    end do

    ! There is no maximum, in other words the smaller number of bins are used,
    ! better the statistics will be.
    if (nDup > 1) then
      if (lowerNbin <= 2) then	! Minimum possible lowerNbin
        if (nDup > sizeMaskIn/3) then
          ! Over one-thirds of the time-series have the same and low Y-values.
          ! Therefore it must be desirable to return the cut-off above the value.
          ! In practice, this must be an unrealistic situation, though.
          write(errStr, '(a,es14.8,a)') 'Failed to find the maximum S/N.  &
             &However, the significant number of lowest bins are found during the search &
             &to have the same counts (or rate): ', curY &
             , '.  They occupy more than one-third of the total number &
             &of bins in the input time-series, therefore this task regards &
             &the value as the cut-off threshold.'
          call warning('regardLowestBinsAsOptimum')	! Then Error is raised.

          optimumY = min(prevY+100*epsilon(prevY), (prevY+prevprevY)/2.0)
          if (present(optimumSize)) optimumSize = count(prevMask)
          if (present(mask)) then
            mask = prevMask
            ! maskOut = prevMask
          end if
          return

        else
          ! Do nothing (ie., proceed)
        end if

      else
        write(errStr, '(a,es14.8)') 'You may want to retry &
           &with a lower mintimeratio (in the command-line argument), &
           &because there is a possibility to find the optimum value with it &
           &given the fact more than one lowest bins are found during the search &
           &to have the same counts (or rate): ', curY
        if (present(status)) then
          call message(AppMsg, Sparse)
        else
          call warning('recommendedLowerMinTimeRatio')	! Then Error is raised.
        end if
      end if
    end if

    if (present(status)) then
      call message(AppMsg, Sparse, 'Failed to find the optimum threshold.')
      status = -2
    else
      call error('noThresholdFound' &
         , 'The smaller number of bins are used, better the statistics will'&
         //' be in the given condition.')
    end if

    if (present(mask)) mask = .false.

  end subroutine findOptimumY


  !---------------------------------------------------------------------
  subroutine calcThresholdFromAny(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status, doDump)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for findOptimumY().
  !  This subroutine adjusts aryY according to fracexp (if that is the case),
  ! makes a mask file, which is false where aryY()<=0 
  ! or where fracexp threshold (if specified so) is not satisfied,
  ! or aryX is out of range in timeStartEnd (if specified so),
  ! and finally passes the variables to findOptimumY().
  ! 
  ! NOTE of caution: If fracexp is given and its %doUse==.true. and
  !  %style="calc", then the returned optimumRate is adjusted accordingly.
  !  For example, if the FRACEXP for every bin is 0.5 and
  ! if the aryY(:) is all 10, then the returned optimumRate will be 20.
  ! In practice, most of FRACEXP is likely to be 1 or very close to 1,
  ! therefore it ie expected this should not cause any practical trouble.
  ! If it ever does, it is up to the user how to use (readjust?) the result.
  !
  ! Arguments:
  !  aryX(:) : TIME column (has to be in order of time.  eg., MJD.  Each time bin must have the same duration.).
  !  aryY(:) : RATE/COUNTS column
  !  optimumRate : The returned Rate value, which is what you want.
  !  optimumSize : The size of the array when the optimum is found.
  !  maskSize    : The size of the mask array before the search.
  !  timeStartEnd, infoFracexp : (OPTIONAL) Constraints if given.
  !    Note the values in timeStartEnd are the actual time (aka value of aryX(i)), NOT the bin number.
  !  lowerCutoffY : (OPTIONAL) The bin with Y equal to or smaller than this value is ignored (The mask is set to be .false.).  This value has to be non-negative.
  !  lowerNbinRatio : (OPTIONAL) Passed to findOptimumY(); In search if the ratio of number of bin currently used to the initial bin number drops below this thoreshold, the calculation (aka search) will stop in findOptimumY().
  !  status : (OPTIONAL) Passed to findOptimumY() as it is; if this is not given and if the search for the minimum fails, then Error is raised.
  ! 

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status
    logical,                 intent(in),  optional :: doDump

    real(single), dimension(size(aryX)) :: tsCnt, aryFracexp
    logical :: mask(size(aryX))
    type(InfoFracexpT) :: localFracexp
    type(ArgTimesStartEndT) :: localTimesStartEnd
    real(single) :: localLowerCutoffY, localLowerNbinRatio !, prevX
    integer(int32) :: nBinIn, i, si

    !-------------------------------------------------------------------
    ! Start

    if (present(optimumSize)) optimumSize = -1
    if (present(maskSize))    maskSize    = -1
    if (present(fracexp))      localFracexp = fracexp
    if (.not. allocated(localFracexp%ary)) allocate(localFracexp%ary(size(aryX)))
    if (present(timeStartEnd)) localTimesStartEnd = timeStartEnd
    localLowerCutoffY = 0.0
    if (present(lowerCutoffY)) localLowerCutoffY = lowerCutoffY
    if (localLowerCutoffY < 0) then
    write(errStr, '(a,es14.8,a)') "lowerCutoffY specified is negative (" &
       , lowerCutoffY &
       , ").  It is reset to zero in calcThresholdFromAny().  Contact the code developer."
      call warning('negativeLowerCutoffY')
    end if
    localLowerNbinRatio = 0.05
    if (present(lowerNbinRatio)) localLowerNbinRatio = lowerNbinRatio

    nBinIn = size(aryX)
    if (nBinIn .ne. size(aryY)) then
      call error('differentSizeOfAryTimeAndRate')
    else if (localFracexp%doUse .and. (nBinIn .ne. size(localFracexp%ary))) then
      call error('differentSizeOfAryTimeAndFracexp')
    end if

    ! Initialisation
    tsCnt = aryY
    ! if (localFracexp%doUse .and. (trim(localFracexp%style) == 'calc')) then
    !   where (localFracexp%ary ==   0) tsCnt = 0
    !   where (localFracexp%ary .ne. 0) tsCnt = aryY / localFracexp%ary
    ! else
    !   tsCnt = aryY
    ! end if

    mask(:) = .true.
    ! prevX = aryX(1) - 1	! Assuming (size(aryX) > 0)
    !! Set the initial value for prevX, which can be arbitrary as long as smaller than aryX(1).

    ! Adjusting the mask array.
    do i=1, nBinIn 
      ! Most of the IF clauses should be located ideally out of the loop
      ! for the sake of the run-time speed.  However they should not be
      ! the bottle neck, so I leave them as they are for readability.

      ! if (prevX >= aryX(i)) then
      !   write(errStr, '(a, i10, a, es16.6, es16.6)') &
      !      'Time bin in the input time series is not in order of time at Bin=' &
      !      , i, ' times:', prevX,  aryX(i)
      !   call error('notSortedTimeBin')
      ! end if

      ! prevX = aryX(i)

      ! if (aryY(i) <= 0) then
      if (.not. (aryY(i) > localLowerCutoffY)) then	! If it is below the threshold (maybe just negative or zero), or NaN.
        mask(i) = .false.
        cycle
      end if

      if (localTimesStartEnd%withStart .and. (aryX(i) < localTimesStartEnd%start)) then
        mask(i) = .false.
        cycle
      end if

      if (localFracexp%doUse .and. (trim(localFracexp%style) == 'threshold')) then
        if (localFracexp%ary(i) < localFracexp%lowerThreshold) then
          mask(i) = .false.
          cycle
        end if
      end if

      if (localTimesStartEnd%withEnd   .and. (aryX(i) > localTimesStartEnd%end)) then
        mask(i) = .false.
        cycle
      end if
    end do		! do i=1, nBinIn 


    si = count(mask)
    if (present(maskSize)) maskSize = si
    write(errStr, '(a,i8,a)') "Mask file has been successfully generated (size="&
       , si, ").  Proceed to calculate."
    call message(AppMsg, Noisy)

    if (localFracexp%doUse .and. (trim(localFracexp%style) == 'calc')) then
      call findOptimumY(tsCnt, optimumRate &
         , optimumSize=optimumSize, aryFracexp=real(localFracexp%ary, single), mask=mask &
         , lowerNbinRatio=localLowerNbinRatio, status=status, doDump=doDump)	! Raise error if goes wrong and if status not given.
    else
      call findOptimumY(tsCnt, optimumRate &
         , optimumSize=optimumSize, mask=mask &
         , lowerNbinRatio=localLowerNbinRatio, status=status, doDump=doDump)	! Raise error if goes wrong and if status not given.
    end if



    deallocate(localFracexp%ary)

  end subroutine calcThresholdFromAny


  !---------------------------------------------------------------------
  subroutine calcThresholdFromRate(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for calcThresholdFromAny().

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status

    real(single) :: aryCnt(size(aryX))
    type(InfoFracexpT) :: localFracexp

    !-------------------------------------------------------------------
    ! Start

    if (present(fracexp)) localFracexp = fracexp
    aryCnt = aryY

    if (fracexp%doUse) then
      if (trim(fracexp%style) == 'calc') then
        aryCnt = 0.0
        where((aryY > 0.0) .and. (fracexp%ary > 0.0)) aryCnt = aryY * fracexp%ary
      end if
    end if

    call calcThresholdFromAny(aryCnt, aryY, optimumRate &
       , optimumSize=optimumSize, maskSize=maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd &
       , lowerCutoffY=lowerCutoffY, lowerNbinRatio=lowerNbinRatio, status=status)
    ! This is not the most efficient, as tsRate, which is aryX(:) here,
    ! will be calculated again in the later stage, subroutine findOptimumY().
    ! However the overhead in practice is probably small enough.

  end subroutine calcThresholdFromRate


  !---------------------------------------------------------------------
  subroutine calcThresholdFromCount(aryX, aryY, optimumRate &
     , optimumSize, maskSize, fracexp, timeStartEnd &
     , lowerCutoffY, lowerNbinRatio, status)
  !---------------------------------------------------------------------
  ! Description:
  !  Wrapper for calcThresholdFromAny().

    implicit none

    real(single), intent(in) :: aryX(:), aryY(:)
    real(single), intent(out) :: optimumRate
    integer(int32),          intent(out), optional :: optimumSize, maskSize
    type(InfoFracexpT),      intent(in),  optional :: fracexp
    type(ArgTimesStartEndT), intent(in),  optional :: timeStartEnd
    real(single),            intent(in),  optional :: lowerCutoffY, lowerNbinRatio
    integer,                 intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    call calcThresholdFromAny(aryX, aryY, optimumRate &
       , optimumSize=optimumSize, maskSize=maskSize &
       , fracexp=fracexp, timeStartEnd=timeStartEnd &
       , lowerCutoffY=lowerCutoffY, lowerNbinRatio=lowerNbinRatio, status=status)

  end subroutine calcThresholdFromCount


end module bkgoptrate_aux
