!
! Module: bkgoptrate_mod
!
! Authors: Masaaki Sakano (SSC/LUX)
!          
! Description: Output the best RATE to maximise the statistics, based on the given background time-series.
! 
! Comments: 
!  * Test with a step-function type time-series.
!   This method should work well, however epsfilt, which uses
!   the Gaussian fitting, may not, because the histogram will be
!   a double-peak function and be far from Gaussian.
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module bkgoptrate_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  ! use cal
  use param
  use errorHandling

  use bkgoptrate_aux

  !! in ssclib/
  ! use array_utils	! for smallest/largestTrueIndex
  ! use constants
  use dal_aux	! for readColToFixed
  ! use ssc_misc	! for intToStr
  ! use minimizations	! for riddersMethod

  implicit none 

  private 

  ! Command-line parameters:

  character(256) ::&
    argBkgTsSetTabName = '',&
    argTsStyle = '',&
    argGtiStyle = '',&
    argColNameX = '',& 
    argColNameY = '',&
    argColNameFracexp = '',&
    argFracexpStyle = ''
    ! srcListSetTabName = '',&
    ! tempSrcSetName = '',&
    ! selectionExpression = '',&
    ! srcRegionSetName = '',&
    ! bkgRegionSetName = '',&
    ! bkgKwdSetName = '',&
    ! bkgImgSetName = ''

  real(single) ::&
    argStartTime = 0.0,&
    argEndTime   = 0.0,&
    argLowerCutoffCount = 0.0,&
    argMinTimeRatio = 0.05	! 5%

  logical ::&
    argWithStartTime = .false.,&
    argWithEndTime   = .false.,&
    argWithLowerCutoffCount = .false.,&
    argWithMinTimeRatio = .false.

  ! Derived from the Command-line parameters:
  character(256) ::&
    BkgTsSetName = '',&
    BkgTsTabName = ''
  character(64) ::&
    TsColNameX = '',&
    TsColNameY = ''
  character(20) ::&
    TsStyle = '',&
    GtiStyle = ''

  type(InfoFracexpT), save :: InfoFracexp		! defined in bkgoptrate_aux
  type(ArgTimesStartEndT), save :: TimesStartEnd	! defined in bkgoptrate_aux

  ! Other globals:

  public bkgoptrate

contains 
 
  !---------------------------------------------------------------------
  subroutine bkgoptrate
  !---------------------------------------------------------------------

    implicit none
    type(DataSetT) :: set
    type(TableT) :: tab
    integer(int32) :: numRows
    ! type(ColumnT) :: colX, colY, colFracExp
    real(single), allocatable :: aryX(:), aryY(:)
    real(single) :: optimumValue
    integer(int32) :: optimumSize, maskSize

logical, parameter :: localTest = .true.

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    !...................................................................
    ! Read the data.

    write(errStr, '(a)') "Read Time-series from (" &
       //trim(BkgTsSetName)//':'//trim(BkgTsTabName)//")."
    call message(AppMsg, Noisy)

    set = dataSet(trim(BkgTsSetName), READ)
    tab = table(set, trim(BkgTsTabName), READ)
    numRows = numberOfRows(tab)
    allocate(aryX(numRows), aryY(numRows))

    call readColDataToFixed(tab, trim(TsColNameX), aryX)
    call readColDataToFixed(tab, trim(TsColNameY), aryY)

    if (InfoFracexp%doUse) then
      allocate(InfoFracexp%ary(numRows))
      call readColDataToFixed(tab, trim(InfoFracexp%colName), InfoFracexp%ary)
    end if

    call release(set)

    write(errStr, '(a)') "Calculating the output threshold..."
    call message(AppMsg, Noisy)

    select case(trim(TsStyle))
    case('rate')
      call calcThresholdFromRate( aryX, aryY, optimumValue &
         , optimumSize=optimumSize, maskSize=maskSize &
         , fracexp=InfoFracexp, timeStartEnd=TimesStartEnd &
         , lowerCutoffY=argLowerCutoffCount, lowerNbinRatio=argMinTimeRatio)

    case('count')
      call calcThresholdFromCount(aryX, aryY, optimumValue &
         , optimumSize=optimumSize, maskSize=maskSize &
         , fracexp=InfoFracexp, timeStartEnd=TimesStartEnd &
         , lowerCutoffY=argLowerCutoffCount, lowerNbinRatio=argMinTimeRatio)

    case default
call error('badTsStyle',"Contact the code developer.")	! Should have been trapped in SUBROUTINE raedParameter().
    end select	! select case(trim(TsStyle))


    !-------- Output --------
    write(errStr, '(i10,a,i10,a,i10,a)') optimumSize &
       , "bins are below the threshold found, out of ", maskSize &
       , "bins after filtering is applied, out of ", size(aryX) &
       , "bins in the original input."
    call message(AppMsg, Noisy)

    write(errStr, '(a)') "The best Y-value is "
    call message(AppMsg, Verbose)

    write(*,'(es13.6)') optimumValue 
    !------------------------


    if (InfoFracexp%doUse) then
      deallocate(InfoFracexp%ary)
    end if

  end subroutine bkgoptrate


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: tsSet
    type(TableT) :: tsTab

    logical :: noColonFound

    !-------------------------------------------------------------------
    ! Start

    argBkgTsSetTabName = stringParameter("tssettabname")
    noColonFound = .false.
    call splitSetTabName(argBkgTsSetTabName, BkgTsSetName, BkgTsTabName &
       , noColonFound=noColonFound)	! in ssclib/dal_aux

    tsSet = dataSet(trim(BkgTsSetName), READ)
    tsTab = table(tsSet, trim(BkgTsTabName))

    argTsStyle = stringParameter("tsstyle")
    TsStyle = trim(argTsStyle)
    argColNameX = stringParameter("xcol")	! TIME (at the Centre) in default
    argColNameY = stringParameter("ycol")

    argGtiStyle = stringParameter("gtistyle")
    argFracexpStyle = stringParameter("fracexpstyle")
    argColNameFracexp = stringParameter("fracexpcol")

    argWithStartTime = booleanParameter("withstarttime")
    if (argWithStartTime) argStartTime = realParameter("starttime")
    argWithEndTime = booleanParameter("withendtime")
    if (argWithEndTime)   argEndTime   = realParameter("endtime")

    argWithLowerCutoffCount = booleanParameter("withlowercutoffcount")
    if (argWithLowerCutoffCount) argLowerCutoffCount = realParameter("lowercutoffcount")
    argWithMinTimeRatio = booleanParameter("withmintimeratio")
    if (argWithMinTimeRatio) argMinTimeRatio = realParameter("mintimeratio")

    ! Determine TsStyle if argTsStyle=='auto'
    if (trim(argTsStyle) == 'auto') then
      if (     hasColumn(tsTab, 'RATE')) then
        TsStyle = 'rate'
        TsColNameY = 'RATE'
      else if (hasColumn(tsTab, 'rate')) then
        TsStyle = 'rate'
        TsColNameY = 'rate'
      else if (hasColumn(tsTab, 'COUNT')) then
        TsStyle = 'count'
        TsColNameY = 'COUNT'
      else if (hasColumn(tsTab, 'count')) then
        TsStyle = 'count'
        TsColNameY = 'count'
      else if (hasColumn(tsTab, 'COUNTS')) then
        TsStyle = 'count'
        TsColNameY = 'COUNTS'
      else if (hasColumn(tsTab, 'counts')) then
        TsStyle = 'count'
        TsColNameY = 'counts'
      else
call error('FailedInTsStyleOfAuto' &
, "Neither 'RATE' nor 'COUNTS?' is found as a column name &
&hence the --tsstyle=auto parameter does not work.")
      end if
    end if


    ! Determine TsColNameX/Y based on --x/ycol.

    if (hasColumn(tsTab, trim(argColNameX))) then
      TsColNameX = trim(argColNameX)
    else
call error('NotExistCountColName' &
, "The specified column ("//trim(argColNameY)//") in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//") does not exist.")
    end if

    select case(trim(TsStyle))
    case('rate')
      if (trim(argColNameY) == 'default') then
        if (     hasColumn(tsTab, 'RATE')) then
          TsColNameY = 'RATE'
        else if (hasColumn(tsTab, 'rate')) then
          TsColNameY = 'rate'
        else
call error('FailedToFindRateColName' &
, "Failed to find the 'RATE' column in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
        end if
      else if (hasColumn(tsTab, trim(argColNameY))) then
        TsColNameY = trim(argColNameY)
      else
call error('NotExistRateColName' &
, "The specified column ("//trim(argColNameY)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case('count')
      if (trim(argColNameY) == 'default') then
        if (     hasColumn(tsTab, 'COUNT')) then
          TsColNameY = 'COUNT'
        else if (hasColumn(tsTab, 'count')) then
          TsColNameY = 'count'
        else if (hasColumn(tsTab, 'COUNTS')) then
          TsColNameY = 'COUNTS'
        else if (hasColumn(tsTab, 'counts')) then
          TsColNameY = 'counts'
        else
call error('FailedToFindCountColName' &
, "Failed to find the 'COUNT' column in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
        end if
      else if (hasColumn(tsTab, trim(argColNameY))) then
        TsColNameY = trim(argColNameY)
      else
call error('NotExistCountColName' &
, "The specified column ("//trim(argColNameY)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case default
call error('badTsStyle',"The value of the --tsstyle parameter is not recognized.")
    end select	! select case(trim(TsStyle))


    write(errStr, '("Column (",a,") is found in (",a,":",a,").")') &
       trim(TsColNameY), trim(BkgTsSetName), trim(BkgTsTabName)
    call message(AppMsg, Noisy)


    select case(trim(argGtiStyle))
    case('auto')
      GtiStyle = 'column'
    case('column')
      GtiStyle = 'column'

    case('gti')
    ! GTI in FITS extentions is generally defined for chip by chip.
    ! Therefore, in order to treat GTI properly, it is necessary
    ! to process like lccorr, which is far too complicated at best,
    ! particularly given lccorr does not work properly (as in 2010)
    ! due to the problem rooted in the DSSLib (fortran interface).
call error('badGtiStyle',"The value of the --gtistyle=gti &
&is not supported at the moment, yet.")

    case default
call error('badGtiStyle',"The value of the --gtistyle parameter &
&is not recognized.")
    end select


    ! Determine InfoFracexp%style if argFracexpStyle=='auto'
    InfoFracexp%doUse = .false.
    if (trim(argFracexpStyle) == 'auto') then
      ! InfoFracexp%style will be either 'calc' or 'none'.
      if (     hasColumn(tsTab, trim(argColNameFracexp))) then
        InfoFracexp%style = 'calc'
        InfoFracexp%colName = trim(argColNameFracexp)
      else
        InfoFracexp%style = 'none'
      end if
    else
      InfoFracexp%style = trim(argFracexpStyle)
    end if

    select case(trim(InfoFracexp%style))
    case('calc', 'threshold')
      if (hasColumn(tsTab, trim(argColNameFracexp))) then
        InfoFracexp%colName = trim(argColNameFracexp)
        InfoFracexp%doUse = .true.

        write(errStr, '("Column (",a,") is found in (",a,":",a,").")') &
           trim(InfoFracexp%colName), trim(BkgTsSetName), trim(BkgTsTabName)
        call message(AppMsg, Noisy)

        if (trim(InfoFracexp%style) == 'threshold') then
          ! The bins, of which FRACEXP is lower than this, are not used.
          InfoFracexp%lowerThreshold = realParameter("fracexplower")
        end if

      else
call error('NotExistFracexpCol' &
, "The specified column ("//trim(argColNameFracexp)//") does not exist in (" &
//trim(BkgTsSetName)//":"//trim(BkgTsTabName)//").")
      end if

    case('none')
        InfoFracexp%doUse = .false.

    case default
call error('badFracexpStyle',"The value of the --fracexpstyle parameter &
&is not recognized.")
    end select


    call release(tsSet)


    TimesStartEnd%withStart = argWithStartTime
    TimesStartEnd%withEnd   = argWithEndTime
    TimesStartEnd%start = argStartTime
    TimesStartEnd%end   = argEndTime

  end subroutine readParameters


!   !---------------------------------------------------------------------
!   function calcThresholdFromCount(aryX, aryY, timeStartEnd, fracexp) result(retThre)
!   !---------------------------------------------------------------------
!   !  aryX: TIME
!   !  aryY: COUNTS
!   ! NOTE: Input arrays are assumed to be sorted in order of the time.
! 
!     implicit none
! 
!     real(single) :: retThre	! Returned value
!     real(double),            intent(in) :: aryX(:), aryY(:)
!     type(ArgTimesStartEndT), intent(in), optional :: timeStartEnd
!     real(double),            intent(in), optional :: fracexp(:)
! 
!     integer(int32) :: numInAry, iStart, iEnd, i, maxAryYLoc(1)
!     logical, allocatable :: aryXMask(:)
!     real(single) :: maxAryY, minAryY, maxAryY10
!     integer :: statusFit
! 
!     !-------------------------------------------------------------------
!     ! Start
! 
!     numInAry = size(aryX) 
!     if (numInAry .ne. size(aryY)) call error('bug', 'Contact the code developer.')
!     if (present(fracexp)) then
!       if (numInAry .ne. size(fracexp)) call error('bug', 'Contact the code developer.')
!     end if
! 
!     ! Determine the first and last indices of X-axis (TIME)
!     iStart = 1
!     iEnd = numInAry
! 
!     if (present(timeStartEnd)) then
!       if (timeStartEnd%withStart) then
!         iStart = numInAry + 1	! Default (Never used in the normal situation)
!         do i=1, numInAry
!           if (aryX(i) >= timeStartEnd%start) then
!             ! This should always be visited at one stage.
!             iStart = i
!             exit
!           end if
!         end do
! 
!         if (iStart > numInAry) then
!           call error('tooLargeStartTime')
!         end if
!       end if	! if (timeStartEnd%withStart) then
! 
!       if (timeStartEnd%withEnd) then
!         iEnd = -1	! Default (Never used in the normal situation)
!         do i=numInAry, 1, -1	! Loop in the reverse order.
!           if (aryX(i) <= timeStartEnd%end) then
!             ! This should always be visited at one stage.
!             iEnd = i
!             exit
!           end if
!         end do
! 
!         if (iEnd < 1) then
!           call error('tooSmallEndTime')
!         end if
!       end if	! if (timeStartEnd%withEnd) then
!     end if	! if (present(timeStartEnd)) then
! 
!     if (iStart == iEnd) then
!       call warning('singleTimeBin', 'After the time-filtering, only a single bin is left.')
!       retThre = valY(iStart)
!       return
!     end if
! 
!     ! Determine the two starting values.
!     allocate(aryXMask(numInAry))
!     aryXMask = .false.
!     aryXMask(iStart:iEnd) = .true.
!     maxAryYLoc = minloc(aryY, aryXMask)	! Actually minimum.
!     minAryY = aryY(maxAryYLoc)
!     maxAryYLoc = maxloc(aryY, aryXMask)
!     maxAryY = aryY(maxAryYLoc)
!     do i=1, int(max(min((iEnd-iStart)/2.0, 10), 1))	! 10-th largest value if defined.
!       aryXMask(maxAryYLoc) = .false.
!       maxAryYLoc = maxloc(aryY, aryXMask)
!       maxAryY10 = aryY(maxAryYLoc)
!     end do
!     deallocate(aryXMask)
! 
!     call riddersMethod(maxAryY, maxAryY10, y, retThre, status=statusFit)	! in ssclib/minimizations.f90
!     !!!! Note that y(startX0) and y(startX1) must be of opposite sign.
!     !!!!!!!!!!!! y is the function y(x)=0
! 
!   end function calcThresholdFromCount


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module bkgoptrate_mod
